var pButton = document.getElementById("prod");
var cButton = document.getElementById("cons");
var bButton = document.getElementById("back");
var countText = document.getElementById("count");
var bufferImg = document.getElementById("bufferpic");
var pStatus = document.getElementById("pstatus");
var cStatus = document.getElementById("cstatus");
var infoDiv = document.getElementById("info");
var infoText = document.getElementById("infotext");
var enableMessages = document.getElementById("enableMessages");

var pButtonImg;
var pButtonImgOver;
var pButtonImgBlocked;
var cButtonImg;
var cButtonImgOver;
var cButtonImgBlocked;
var bButtonImg;
var bButtonImgOver;
var infoIcon = "<img style=\"margin:0px 5px 0px 5px;float:left;\" src=\"imgs/info.png\"></img>";

var COUNT_STRING = "count: ";
var STATUS_READY = 0;
var STATUS_BLOCKED = 1;
var STATUS_EXECUTING = 2;

var STATUS_READY_STR;
var STATUS_BLOCKED_STR;
var STATUS_EXECUTING_STR;

var INFO_TEXT_INCONSISTENT_COUN;
var INFO_TEXT_BOTH_SLEEPING;
var INFO_TEXT_SIGNAL_LOSS_IMMINENT_P;
var INFO_TEXT_SIGNAL_LOSS_IMMINENT_C;

function setLang(lang)
{
	switch(lang)
	{
		case 0:
			STATUS_READY_STR = "Ready";
			STATUS_BLOCKED_STR = "Blocked";
			STATUS_EXECUTING_STR = "Executing";
			INFO_TEXT_INCONSISTENT_COUNT = "At this moment the value of the variable <i>count</i> is inconsistent, since it differs from the amount of items actually in the buffer. This happens every time there's exactly one process stopped at the line that updates <i>count</i>. This, along with the lack of mutual exclusion, may cause little bugs: the producer may block when there's still an empty position in the buffer, and the consumer may do it when there's still an item to be removed.";
			INFO_TEXT_BOTH_SLEEPING = "Due to the loss of a call to <i>wakeup</i>, the producer and the consumer are now blocked potentially forever.";
			INFO_TEXT_SIGNAL_LOSS_IMMINENT_C = "If the consumer is executed now, a <i>wakeup</i> call is going to be lost."
			INFO_TEXT_SIGNAL_LOSS_IMMINENT_P = "If the producer is executed now, a <i>wakeup</i> call is going to be lost."
			pButtonImg = "imgs/steppro_en.png";
			pButtonImgOver = "imgs/steppro_b_en.png";
			pButtonImgBlocked = "imgs/nopro_en.png";
			cButtonImg = "imgs/stepcon_en.png";
			cButtonImgOver = "imgs/stepcon_b_en.png";
			cButtonImgBlocked = "imgs/nocon_en.png";
			bButtonImg = "imgs/stepback_en.png";
			bButtonImgOver = "imgs/stepback_b_en.png";
			
			break;
			
		case 1:
		default:
			STATUS_READY_STR = "Pronto";
			STATUS_BLOCKED_STR = "Bloqueado";
			STATUS_EXECUTING_STR = "Executando";
			INFO_TEXT_INCONSISTENT_COUNT = "Neste momento o valor da variável <i>count</i> está inconsistente, pois difere da quantidade de itens de fato no buffer. Isso acontece sempre que há exatamente um processo parado na linha que atualiza <i>count</i>. Aliado à falta de exclusão mútua, isso pode causar <i>bugs</i> minoritários: o produtor pode bloquear quando ainda há uma posição vaga no <i>buffer</i>, e o consumidor pode fazê-lo quando ainda há um item a ser retirado.";
			INFO_TEXT_BOTH_SLEEPING = "Por causa da perda de uma chamada a <i>wakeup</i>, o produtor e o consumidor agora estão bloqueados potencialmente para sempre.";
			INFO_TEXT_SIGNAL_LOSS_IMMINENT_C = "Se o consumidor for executado agora, uma chamada a <i>wakeup</i> será perdida."
			INFO_TEXT_SIGNAL_LOSS_IMMINENT_P = "Se o produtor for executado agora, uma chamada a <i>wakeup</i> será perdida."
			pButtonImg = "imgs/steppro.png";
			pButtonImgOver = "imgs/steppro_b.png";
			pButtonImgBlocked = "imgs/nopro.png";
			cButtonImg = "imgs/stepcon.png";
			cButtonImgOver = "imgs/stepcon_b.png";
			cButtonImgBlocked = "imgs/nocon.png";
			bButtonImg = "imgs/stepback.png";
			bButtonImgOver = "imgs/stepback_b.png";
			
			break;
	}
}

var buffer = {
    capacity: 5, 
    currSize: 0,
	actualSize: 0,
    isFull: function() {return this.capacity == this.currSize;},
    isEmpty: function() {return this.currSize == 0;},
    inc: function() 
    {
        if (!(this.isFull()))
        {
            this.currSize++;
            countText.innerHTML = COUNT_STRING + this.currSize;
        }
    },
    dec: function() 
    {
        if (!(this.isEmpty()))
        {
            this.currSize--;
            countText.innerHTML = COUNT_STRING + this.currSize;
        }
    },
	insert: function()
	{
		if(this.actualSize >= this.capacity)
			return false;
		
		this.actualSize++;
		bufferImg.src = "imgs/" + this.actualSize + ".png";
		return true;
	},
	remove: function()
	{
		if (this.actualSize <= 0)
			return false;
		
		this.actualSize--;
		bufferImg.src = "imgs/" + this.actualSize + ".png";
	}
};

var producer = {
    sleeping: false,
    currLine: 1,
    sleep: function()
    {
        //pButton.disabled = true;
        //pButton.innerHTML = "Step Producer (sleeping)";
		pButton.src = pButtonImgBlocked;
        this.setStatus(STATUS_BLOCKED);
        this.sleeping = true;
    },
    wakeup: function()
    {
        //pButton.disabled = false;
        //pButton.innerHTML = "Step Producer";
		pButton.src = pButtonImg;
        this.setStatus(STATUS_READY);
        this.sleeping = false;
    },
	status: STATUS_READY,
    getStatus: function()
    {
        return this.status;
    },
    setStatus: function(status)
    {
		switch(status)
		{
			case STATUS_READY:
				pStatus.innerHTML = STATUS_READY_STR;
				pStatus.style.color = "black";
				break;
				
			case STATUS_BLOCKED:
				pStatus.innerHTML = STATUS_BLOCKED_STR;
				pStatus.style.color = "red";
				break;
				
			case STATUS_EXECUTING:
				pStatus.innerHTML = STATUS_EXECUTING_STR;
				pStatus.style.color = "green";
				break;
		}
		this.status = status;
    },
    step: function()
    {	
        this.setStatus(STATUS_EXECUTING);
        
        if(consumer.getStatus() == STATUS_EXECUTING)
            consumer.setStatus(STATUS_READY);
            
        var nextLine;
        
        switch(this.currLine)
        {
            case 1:
                nextLine = this.currLine + 1;
                break;
            
            case 2:
                nextLine = this.currLine + 1;
                break;
                
            case 3:
                if (buffer.isFull())
                    nextLine = this.currLine + 1;
                else
                    nextLine = this.currLine + 2;
                break;

            case 4:
                this.sleep();
                nextLine = this.currLine + 1;
                break;
                
            case 5:
				buffer.insert();
                nextLine = this.currLine + 1;
                break;
                
            case 6:
                buffer.inc();
                nextLine = this.currLine + 1;
                break;
                
            case 7:
                if (buffer.currSize == 1)
                    nextLine = this.currLine + 1;
                else
                    nextLine = 1;
                break;
                    
            case 8:
                consumer.wakeup();
                nextLine = 1;
                break;
        }

        this.setLine(nextLine);
    },
    setLine: function(line) 
    {
        var oldLine = this.currLine;
        var oldLineId = "pl" + oldLine;
        var lineId;
        
        var lineId = "pl" + line;
        document.getElementById(oldLineId).className = "codeline";
        document.getElementById(lineId).className = "codeline selected";
        this.currLine = line;
    }
};
    
var consumer = {
    sleeping: false,
    currLine: 1,
    sleep: function()
    {
        //cButton.disabled = true;
        //cButton.innerHTML = "Step Consumer (sleeping)";
		cButton.src = cButtonImgBlocked;
        this.setStatus(STATUS_BLOCKED);
        this.sleeping = true;
    },
    wakeup: function()
    {
        //cButton.disabled = false;
        //cButton.innerHTML = "Step Consumer";
		cButton.src = cButtonImg;
        this.setStatus(STATUS_READY);
        this.sleeping = false;
    },
	status: STATUS_READY,
    getStatus: function()
    {
        return this.status;
    },
    setStatus: function(status)
    {
		switch(status)
		{
			case STATUS_READY:
				cStatus.innerHTML = STATUS_READY_STR;
				cStatus.style.color = "black";
				break;
				
			case STATUS_BLOCKED:
				cStatus.innerHTML = STATUS_BLOCKED_STR;
				cStatus.style.color = "red";
				break;
				
			case STATUS_EXECUTING:
				cStatus.innerHTML = STATUS_EXECUTING_STR;
				cStatus.style.color = "green";
				break;
		}
		this.status = status;
    },
    step: function()
    {   
        this.setStatus(STATUS_EXECUTING);

        if(producer.getStatus() == STATUS_EXECUTING)
            producer.setStatus(STATUS_READY);
            
        var nextLine;
            
        switch(this.currLine)
        {
            case 1:
                nextLine = this.currLine+1;
                break;
            
            case 2:
                if(buffer.isEmpty())
                    nextLine = this.currLine+1;
                else
                    nextLine = this.currLine+2;
                break;
                
            case 3:
                this.sleep();
                nextLine = this.currLine+1;
                break;

            case 4:
				buffer.remove();
                nextLine = this.currLine+1;
                break;
                
            case 5:
                nextLine = this.currLine+1;
                buffer.dec();
                break;
                
            case 6:
                if (buffer.currSize == buffer.capacity - 1)
                    nextLine = this.currLine+1;
                else
                    nextLine = this.currLine+2;
                break;
                
            case 7:
                producer.wakeup();
                nextLine = this.currLine+1;
                break;
                    
            case 8:
                nextLine = 1;
                break;
            
        }

        this.setLine(nextLine);
    },
    setLine: function(line) 
    {
        var oldLine = this.currLine;
        var oldLineId = "cl" + oldLine, lineId;
        
        var lineId = "cl" + line;
        document.getElementById(oldLineId).className = "codeline";
        document.getElementById(lineId).className = "codeline selected";
        this.currLine = line;
    }
};
    
function Context(producerLine, consumerLine, bufferSize, actualBufferSize)
{
    this.pLine = producerLine;
    this.cLine = consumerLine;
    this.bSize = bufferSize;
	this.abSize = actualBufferSize;
    
    this.load = function()
    {
        if (this.pLine < 0)
        {
            this.pLine = -this.pLine;
            producer.sleep();
        }
        else
            producer.wakeup();
        
        if (this.cLine < 0)
        {
            this.cLine = -this.cLine;
            consumer.sleep();
        }
        else
            consumer.wakeup();
            
        consumer.setLine(this.cLine);
        producer.setLine(this.pLine);
        buffer.currSize = this.bSize;
		buffer.actualSize = this.abSize;
        countText.innerHTML = COUNT_STRING + this.bSize;
		bufferImg.src = "imgs/" + this.abSize + ".png";
    }
}

var contextArray = [];

var contextManager = {
    
    getContext: function()
    {
        var cLine = consumer.currLine;
        if (consumer.sleeping == true)
            cLine = -cLine;
            
        var pLine = producer.currLine;
        if (producer.sleeping == true)
            pLine = -pLine;
        
        var currentContext = new Context(pLine, cLine, buffer.currSize, buffer.actualSize);
        return currentContext;
    },
    stepConsumer: function()
    {
		if (consumer.getStatus() == STATUS_BLOCKED)
			return;
		
        contextArray.push(this.getContext());
        consumer.step();
        updateInfo();
    },  
    stepProducer: function()
    {
		if(producer.getStatus() == STATUS_BLOCKED)
			return;
	
        contextArray.push(this.getContext());
        producer.step();
        updateInfo();
    },
    stepBack: function()
    {
        if (contextArray.length > 0)
        {
			(contextArray.pop()).load();
			updateInfo();
        }
    },
	saveSession: function()
	{
		contextArray.push(this.getContext());
		sessionStorage.contextArray = JSON.stringify(contextArray);
		contextArray.pop();
	},
	restoreSession: function()
	{
		if (typeof sessionStorage.contextArray != "undefined")
		{
			contextArray = JSON.parse(sessionStorage.contextArray);
			var c = new Context(-1, -1, -1, -1);
			
			for (i = 0; i < contextArray.length; i++)
				contextArray[i].load = c.load;
			
			(contextArray.pop()).load();
			updateInfo();
		}
	}
};

function blink(buttonId)
{
	switch(buttonId)
	{
		case 1:
			if(producer.getStatus() != STATUS_BLOCKED)
				pButton.src = pButtonImgOver;
			break;
			
		case -1:
			if(producer.getStatus() != STATUS_BLOCKED)
				pButton.src = pButtonImg;
			break;
			
		case 2:
			if(consumer.getStatus() != STATUS_BLOCKED)
				cButton.src = cButtonImgOver;
			break;
			
		case -2:
			if(consumer.getStatus() != STATUS_BLOCKED)
				cButton.src = cButtonImg;
			break;
			
					
		case 3:
			bButton.src = bButtonImgOver;
			break;
			
		case -3:
			bButton.src = bButtonImg;
			break;
	}
}

function displayMessage(message)
{
	infoDiv.style.display = "block";
	infoText.innerHTML = infoIcon + message;
}

function hideMessage()
{
	infoDiv.style.display = "none";
	infoText.innerHTML = "";
}

function updateInfo()
{
	if (!enableMessages.checked)
	{
		hideMessage();
		return;
	}
	
	if (buffer.currSize != buffer.actualSize)
		displayMessage(INFO_TEXT_INCONSISTENT_COUNT);
	
	else if (consumer.sleeping == true && producer.sleeping == true)
		displayMessage(INFO_TEXT_BOTH_SLEEPING);
	
	else if(producer.currLine == 8 && !consumer.sleeping)
		displayMessage(INFO_TEXT_SIGNAL_LOSS_IMMINENT_P);
	
	else if(consumer.currLine == 7 && !producer.sleeping)
		displayMessage(INFO_TEXT_SIGNAL_LOSS_IMMINENT_C);
	
	else
		hideMessage();
}
